from typing import Iterator

from django.db import transaction
from django.db.models import F

from statismo.collectors.base import BaseCollector
from statismo.metrics import Value, Dimension
from statismo.models import Metric, Unit


class DatabaseCollector(BaseCollector):
    def metrics(self) -> Iterator[Metric]:
        """
        Returns an iterator over all available metrics.
        """
        with transaction.atomic():
            q = Metric.objects.select_for_update()

            for metric in q:
                yield metric

            q.update(value=0)

    def save(
        self,
        key: str,
        value: Value,
        unit: Unit,
        dimhash: str,
        *,
        dimensions: list[Dimension] = None,
    ):
        """
        Save a metric to the collector.
        """
        metric, created = Metric.objects.get_or_create(
            key=key,
            dimensions_hash=dimhash,
            defaults={"value": value, "unit": unit, "dimensions": dimensions},
        )
        if not created:
            metric.value = F("value") + value
            metric.save(update_fields=["value"])



import decimal
import enum
import dataclasses
import hashlib
import json
from functools import cache

from django.conf import settings
from django.utils.module_loading import import_string

from statismo.models import Unit


class MetricDestination(enum.Enum):
    DEFAULT = "default"


Value = float | decimal.Decimal | int


@dataclasses.dataclass
class Dimension:
    name: str
    value: str | Value


@cache
def collector():
    return import_string(
        getattr(
            settings,
            "STATISMO_METRIC_COLLECTOR",
            "statismo.collectors.db.DatabaseCollector",
        )
    )()


def push_metric(
    key: str,
    value: Value,
    unit: Unit,
    *,
    dimensions: list[Dimension] = None,
):
    # We use ":" as a divider cache key names, and cloudwatch itself
    # forbids ":" in metric names.
    if ":" in key:
        raise ValueError("Metric key cannot contain ':'")

    dims = []
    for dimension in dimensions or []:
        if ":" in dimension.name:
            raise ValueError("Dimension name cannot contain ':'")

        dims.append({"name": dimension.name, "value": dimension.value})

    dims_hash = hashlib.md5()
    dims_hash.update(json.dumps(dims, sort_keys=True).encode("utf-8"))

    collector().save(
        key=key,
        value=value,
        unit=unit,
        dimhash=dims_hash.hexdigest(),
        dimensions=dims,
    )



import abc
from typing import Iterable

from django.conf import settings

from statismo.metrics import Value, Dimension
from statismo.models import Unit, Metric


class BaseCollector(abc.ABC):
    def metrics(self) -> Iterable[Metric]:
        """
        Returns an iterator over all available metrics.
        """

    def save(
        self,
        key: str,
        value: Value,
        unit: Unit,
        dimhash: str,
        *,
        dimensions: list[Dimension] = None,
    ):
        """
        Save a metric to the collector.
        """

    @staticmethod
    def get_setting(name, default=None):
        """
        Get a setting from the Django settings.

        :param name: The name of the setting.
        :param default: The default value to return if the setting is not set.
        :return: The value of the setting.
        """
        return getattr(settings, f"STATISMO_METRIC_{name}", default)



import pytest
from unittest.mock import patch, MagicMock

from statismo.models import Metric
from statismo.tasks import accumulate_metrics

@pytest.mark.django_db
def test_accumulate_metrics():
    """Test the accumulate_metrics task that sends metrics to CloudWatch."""
    # Create some test metrics
    Metric.objects.create(
        key="test_metric1",
        value=42.0,
        unit="20",  # Unit.COUNT_PER_SECOND as string
        dimensions=[{"name": "service", "value": "test"}],
        dimensions_hash="hash1",
    )

    Metric.objects.create(
        key="test_metric2",
        value=100.0,
        unit="10",  # Unit.SECONDS as string
        dimensions=[{"name": "service", "value": "test2"}],
        dimensions_hash="hash2",
    )

    # Mock boto3 client
    mock_client = MagicMock()

    # Replace the unit mapping function to avoid patching the enum
    unit_get_mapping = MagicMock(return_value="Count/Second")

    with (
        patch("boto3.client", return_value=mock_client),
        patch("statismo.tasks.settings.DEBUG", False),
    ):
        # Run the task
        accumulate_metrics()

        # Verify boto3 client was called with correct data
        mock_client.put_metric_data.assert_called_once()

        # Get the call arguments
        args, kwargs = mock_client.put_metric_data.call_args

        # Verify namespace
        assert kwargs["Namespace"] == "statismo"

        # Verify metrics data
        metrics_data = kwargs["MetricData"]
        assert len(metrics_data) == 2

        # Convert to a dict for easier testing
        metrics_by_name = {m["MetricName"]: m for m in metrics_data}

        # Check first metric
        assert "test_metric1" in metrics_by_name
        m1 = metrics_by_name["test_metric1"]
        assert m1["Value"] == 42.0
        # Skip unit test since that depends on the mapping
        assert len(m1["Dimensions"]) == 1
        assert m1["Dimensions"][0]["Name"] == "service"
        assert m1["Dimensions"][0]["Value"] == "test"

        # Check second metric
        assert "test_metric2" in metrics_by_name
        m2 = metrics_by_name["test_metric2"]
        assert m2["Value"] == 100.0
        # Skip unit test since that depends on the mapping
        assert len(m2["Dimensions"]) == 1
        assert m2["Dimensions"][0]["Name"] == "service"
        assert m2["Dimensions"][0]["Value"] == "test2"

        # Verify metrics were reset
        for metric in Metric.objects.all():
            assert metric.value == 0



# Generated by Django 4.1.7 on 2023-09-18 17:48

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Metric",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=255)),
                ("value", models.DecimalField(decimal_places=4, max_digits=10)),
                ("dimensions", models.JSONField(default=list)),
                (
                    "dimensions_hash",
                    models.CharField(max_length=255, null=True),
                ),
                (
                    "unit",
                    models.CharField(
                        choices=[
                            (10, "Seconds"),
                            (20, "Count/Second"),
                            (30, "None"),
                        ],
                        max_length=255,
                    ),
                ),
            ],
        ),
        migrations.AddIndex(
            model_name="metric",
            index=models.Index(fields=["key"], name="metrics_key_9782fc_idx"),
        ),
        migrations.AddConstraint(
            model_name="metric",
            constraint=models.UniqueConstraint(
                fields=("key", "dimensions_hash"), name="unique_metric"
            ),
        ),
    ]



from django.contrib import admin

from statismo.models import Metric


# Register your models here.
@admin.register(Metric)
class MetricAdmin(admin.ModelAdmin):
    list_display = (
        "key",
        "value",
        "unit",
    )
    list_filter = ("key",)
    search_fields = ("key",)



SECRET_KEY = "django-insecure-test-key"
DEBUG = True

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "statismo",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

USE_TZ = True



import pytest
from unittest.mock import patch, MagicMock

from statismo.metrics import push_metric, Dimension
from statismo.models import Unit, Metric
from statismo.collectors.db import DatabaseCollector
from statismo.collectors.redis import RedisCollector


@pytest.mark.django_db
def test_push_metric_with_db_collector():
    """Test that push_metric correctly stores data in the database."""
    with patch("statismo.metrics.collector", return_value=DatabaseCollector()):
        push_metric(
            key="test_metric",
            value=42.0,
            unit=Unit.COUNT_PER_SECOND,
            dimensions=[Dimension(name="service", value="test")],
        )

        metric = Metric.objects.get(key="test_metric")
        assert metric.value == 42.0
        assert metric.unit == "20"  # Unit.COUNT_PER_SECOND value as string
        assert len(metric.dimensions) == 1
        assert metric.dimensions[0]["name"] == "service"
        assert metric.dimensions[0]["value"] == "test"

        push_metric(
            key="test_metric",
            value=8.0,
            unit=Unit.COUNT_PER_SECOND,
            dimensions=[Dimension(name="service", value="test")],
        )

        metric.refresh_from_db()
        assert metric.value == 50.0


@pytest.mark.django_db
@pytest.mark.redis
def test_push_metric_with_redis_collector():
    """Test that push_metric correctly stores data in Redis using a mock."""
    # Skip if redis is not installed
    pytest.importorskip("redis")

    # Mock Redis client
    mock_redis = MagicMock()
    mock_redis.keys.return_value = [b"statismo:metrics:test_metric:hash123"]
    mock_redis.hgetall.return_value = {
        b"value": b"42.0",
        b"unit": b"20",  # Unit.COUNT_PER_SECOND value
        b"dimensions": b'[{"name": "service", "value": "test"}]',
    }

    class MockRedisCollector(RedisCollector):
        def __init__(self):
            self.client = mock_redis

        def get_setting(self, name):
            return "redis://localhost:6379/0"

    with patch("statismo.metrics.collector", return_value=MockRedisCollector()):
        push_metric(
            key="test_metric",
            value=42.0,
            unit=Unit.COUNT_PER_SECOND,
            dimensions=[Dimension(name="service", value="test")],
        )

        # Verify metric was saved to Redis
        mock_redis.hincrbyfloat.assert_called_with(
            f"statismo:metrics:test_metric:"
            f"{mock_redis.hincrbyfloat.call_args[0][0].split(':')[-1]}",
            "value",
            42.0,
        )
        mock_redis.hset.assert_any_call(
            mock_redis.hincrbyfloat.call_args[0][0],
            "unit",
            Unit.COUNT_PER_SECOND.value,
        )

        metrics = list(MockRedisCollector().metrics())
        assert len(metrics) == 1
        assert metrics[0].key == "test_metric"
        assert metrics[0].value == 42.0
        assert metrics[0].unit == Unit.COUNT_PER_SECOND
        assert len(metrics[0].dimensions) == 1
        assert metrics[0].dimensions[0]["name"] == "service"
        assert metrics[0].dimensions[0]["value"] == "test"

        mock_redis.hset.assert_any_call(
            mock_redis.keys.return_value[0], "value", 0
        )



from django.db import models


class Unit(models.IntegerChoices):
    SECONDS = (10, "Seconds")
    COUNT_PER_SECOND = (20, "Count/Second")
    NONE = (30, "None")


class Metric(models.Model):
    """
    Simple database-backed metric accumulator.

    .. note::

        Not at all suitable for high-volume metrics, but good enough for
        basic usage. Make sure to vacuum the table frequently.
    """

    key = models.CharField(max_length=255)
    value = models.DecimalField(decimal_places=4, max_digits=10)
    dimensions = models.JSONField(default=list)
    dimensions_hash = models.CharField(max_length=255, null=True)
    unit = models.CharField(choices=Unit.choices, max_length=255)

    class Meta:
        indexes = [
            models.Index(fields=["key"]),
        ]
        constraints = [
            models.UniqueConstraint(
                name="statismo_unique_metric",
                fields=["key", "dimensions_hash"],
            ),
        ]



import itertools

import boto3
from celery import shared_task
from django.db import transaction
from django.conf import settings

from statismo.models import Unit
from statismo.metrics import collector


def chunker(it, size):
    """
    Chunk an iterator into chunks of size `size`.

    If the final chunk is smaller than `size`, it will be returned as-is
    without padding.

    :param it: The iterator to chunk.
    :param size: The size of each chunk.
    :return: An iterator of chunks.
    """
    iterator = iter(it)
    while chunk := list(itertools.islice(iterator, size)):
        yield chunk


@shared_task(ignore_result=True)
def accumulate_metrics():
    """
    Accumulate metrics from the database.

    This task runs periodically to accumulate metrics from the metric collector
    and send them to CloudWatch.
    """
    if settings.DEBUG:
        return

    with transaction.atomic():
        metrics = collector().metrics()

        # Maximum # of metrics is 150 per call. Maximum payload size 1MB.
        for chunk in chunker(metrics, 75):
            chunk = list(chunk)

            client = boto3.client("cloudwatch", region_name="us-west-2")
            client.put_metric_data(
                Namespace=getattr(
                    settings, "STATISMO_METRIC_NAMESPACE", "statismo"
                ),
                MetricData=[
                    {
                        "MetricName": m.key,
                        "Dimensions": [
                            {
                                "Name": d["name"],
                                "Value": d["value"],
                            }
                            for d in (m.dimensions or [])
                        ],
                        "Value": m.value,
                        "Unit": {
                            Unit.SECONDS: "Seconds",
                            Unit.COUNT_PER_SECOND: "Count/Second",
                            Unit.NONE: "None",
                        }.get(m.unit, "None"),
                    }
                    for m in chunk
                ],
            )



import json
from typing import Iterator

import redis

from statismo.collectors.base import BaseCollector
from statismo.metrics import Dimension, Value
from statismo.models import Unit, Metric


class RedisCollector(BaseCollector):
    """
    Stores metrics in Redis.
    """

    def __init__(self):
        self.client = redis.from_url(self.get_setting("URL"))

    def metrics(self) -> Iterator[Metric]:
        """
        Returns an iterator over all available metrics.

        Each value is reset to 0 after being read.
        """
        metric_keys = self.client.keys("statismo:metrics:*")
        for key in metric_keys:
            result = self.client.hgetall(key)
            yield Metric(
                key=key.decode("utf-8").split(":")[2],
                value=float(result[b"value"]),
                unit=Unit(int(result[b"unit"])),
                dimensions=(
                    json.loads(result[b"dimensions"])
                    if result.get(b"dimensions")
                    else None
                ),
            )
            self.client.hset(key, "value", 0)

    def save(
        self,
        key: str,
        value: Value,
        unit: Unit,
        dimhash: str,
        *,
        dimensions: list[Dimension] = None,
    ):
        """
        Save a metric to the collector.

        Increments the existing value, if any.
        """
        key = f"statismo:metrics:{key}:{dimhash}"
        self.client.hincrbyfloat(key, "value", value)
        self.client.hset(key, "unit", unit.value)
        if dimensions:
            self.client.hset(key, "dimensions", json.dumps(dimensions))
